package de.ozgcloud.admin.setting;

import static org.assertj.core.api.Assertions.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import java.util.Map;
import java.util.Optional;

import org.apache.commons.lang3.StringUtils;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Spy;
import org.springframework.cloud.config.environment.Environment;
import org.springframework.cloud.config.environment.PropertySource;

import com.fasterxml.jackson.annotation.JsonUnwrapped;
import com.fasterxml.jackson.databind.ObjectMapper;

import lombok.Getter;

class SettingEnvironmentRepositoryTest {
    @InjectMocks
    @Spy
    private SettingEnvironmentRepository repository;
    @Spy
    private ObjectMapper objectMapper;
    @Mock
    private SettingService service;
    @Mock
    private ApplicationSettingDTO mockedSettingDTO;
    @Mock
    private Environment mockedEnvironment;
    @Mock
    private PropertySource mockedPropertySource;

    private static final String ALFA = "Alfa";
    private static final String VORGANG_MANAGER = "OzgCloud_VorgangManager";

    @Nested
    class TestFindOne {
        private static final String ANY_APPLICATION = "any application";

        @Test
        void shouldCallGetAnwendungsSettingDTO() {
            repository.findOne(ANY_APPLICATION, "any profile", "any label");

            verify(repository).findAnwendungSettingDTO(ANY_APPLICATION);

        }

        @Test
        void shouldCallBuildEnvironment() {
            doReturn(Optional.of(mockedSettingDTO)).when(repository).findAnwendungSettingDTO(any());
            doReturn(mockedEnvironment).when(repository).buildEnvironment(any(), any());

            repository.findOne(ANY_APPLICATION, "any profile", "any label");

            verify(repository).buildEnvironment(ANY_APPLICATION, Optional.of(mockedSettingDTO));
        }

        @Test
        void shouldReturnBuiltEnvironment() {
            doReturn(Optional.of(mockedSettingDTO)).when(repository).findAnwendungSettingDTO(any());
            doReturn(mockedEnvironment).when(repository).buildEnvironment(any(), any());

            var returnedEnvironment = repository.findOne(ANY_APPLICATION, "any profile", "any label");

            assertThat(returnedEnvironment).isEqualTo(mockedEnvironment);
        }

    }

    @Nested
    class TestFindAnwendungsSettingDTO {

        @Test
        void shouldReturnEmptyOptionalForUnsupportedApplication() {
            var returnedSettingDTO = repository.findAnwendungSettingDTO("UnsupportedApp");

            assertThat(returnedSettingDTO).isEmpty();
        }

        @Nested
        class TestWithAlfaApplication {
            @Mock
            private AlfaSettingDTO mockedAlfaSettingDTO;

            @BeforeEach
            void mockService() {
                when(service.getAlfaSettingDTO()).thenReturn(mockedAlfaSettingDTO);
            }

            @Test
            void shouldOnlyCallGetAlfaSettingDTO() {
                repository.findAnwendungSettingDTO(ALFA);

                verify(service).getAlfaSettingDTO();
                verify(service, never()).getVorgangManagerSettingDTO();
            }

            @Test
            void shouldReturnAlfaSettingDTO() {
                var returnedSettingDTO = repository.findAnwendungSettingDTO(ALFA);

                assertThat(returnedSettingDTO).hasValue(mockedAlfaSettingDTO);
            }
        }

        @Nested
        class TestWithVorgangManagerApplication {
            @Mock
            private VorgangManagerSettingDTO mockedVorgangManagerSettingDTO;

            @BeforeEach
            void mockService() {
                when(service.getVorgangManagerSettingDTO()).thenReturn(mockedVorgangManagerSettingDTO);
            }

            @Test
            void shouldOnlyCallGetVorgangManagerSettingDTO() {
                repository.findAnwendungSettingDTO(VORGANG_MANAGER);

                verify(service, never()).getAlfaSettingDTO();
                verify(service).getVorgangManagerSettingDTO();
            }

            @Test
            void shouldReturnAlfaSettingDTO() {
                var returnedSettingDTO = repository.findAnwendungSettingDTO(VORGANG_MANAGER);

                assertThat(returnedSettingDTO).hasValue(mockedVorgangManagerSettingDTO);
            }
        }
    }

    @Nested
    class TestBuildEnvironment {
        private static final String APPLICATION_NAME = "app";

        @BeforeEach
        void mockTransformToPropertySource() {
            doReturn(mockedPropertySource).when(repository).transformToPropertySource(mockedSettingDTO);
        }

        @Test
        void shouldCallTransformToPropertySource() {
            repository.buildEnvironment(APPLICATION_NAME, Optional.of(mockedSettingDTO));

            verify(repository).transformToPropertySource(mockedSettingDTO);
        }

        @Test
        void shouldReturnBuiltEnvironment() {
            var expectedEnvironemnt = new Environment(APPLICATION_NAME);
            expectedEnvironemnt.add(mockedPropertySource);

            var resultEnvironment = repository.buildEnvironment(APPLICATION_NAME, Optional.of(mockedSettingDTO));

            assertThat(resultEnvironment).usingRecursiveComparison().isEqualTo(expectedEnvironemnt);
        }
    }

    @Nested
    class TestTransformToPropertySource {
        private ApplicationSettingDTO flatSettingDTO = new ApplicationSettingDTO() {
            @Getter
            private String attribute = "outerValue";
        };

        private ApplicationSettingDTO nestedSettingDTO = new ApplicationSettingDTO() {
            @JsonUnwrapped(prefix = "attribute.")
            @Getter
            private ApplicationSettingDTO attribute = new ApplicationSettingDTO() {
                @Getter
                private String nestedAttribute = "innerValue";
            };
        };

        @Test
        void shouldReturnPropertySourceOfFlatObject() {
            var expectedPropertySource = new PropertySource(StringUtils.EMPTY, Map.of("attribute", "outerValue"));

            var resultPropertySource = repository.transformToPropertySource(flatSettingDTO);

            assertThat(resultPropertySource).usingRecursiveComparison().isEqualTo(expectedPropertySource);
        }

        @Test
        void shouldReturnPropertySourceOfNestedObject() {
            var expectedPropertySource = new PropertySource(StringUtils.EMPTY, Map.of("attribute.nestedAttribute", "innerValue"));

            var resultPropertySource = repository.transformToPropertySource(nestedSettingDTO);

            assertThat(resultPropertySource).usingRecursiveComparison().isEqualTo(expectedPropertySource);
        }
    }
}
